#include <iostream>
#include <span>
#include <vector>

#include "reduce.cuh"

auto divRoundUp(auto a, auto b)
{
    return (a + b - 1) / b;
}

int main()
{
    constexpr size_t size = 1 << 25;
    using type            = float;

    constexpr unsigned block_size       = 1 << 11; // Too large, max is 1<<10
    constexpr unsigned chunks_per_block = 1;

    const auto vec_host  = std::vector< type >(size, 1);
    const auto span_host = std::span{vec_host};

    type *vec1_device{}, *vec2_device{};
    cudaMalloc((void**)&vec1_device, span_host.size_bytes());
    cudaMalloc((void**)&vec2_device, divRoundUp(span_host.size(), block_size * chunks_per_block) * sizeof(type));
    cudaMemcpyAsync(vec1_device, span_host.data(), span_host.size_bytes(), cudaMemcpyHostToDevice);

    for (size_t sz = size;;)
    {
        const unsigned next_sz = divRoundUp(sz, block_size * chunks_per_block);
        reduction< block_size, chunks_per_block ><<< next_sz, block_size >>>(vec1_device, sz, vec2_device);
        std::swap(vec1_device, vec2_device);
        sz = next_sz;
        if (sz == 1)
            break;
    }

    type result = 0.;
    cudaMemcpyAsync(&result, vec1_device, sizeof(type), cudaMemcpyDeviceToHost);

    cudaDeviceSynchronize();
    std::cout << "Reduction result: " << result << '\n';

    cudaFree(vec1_device);
    cudaFree(vec2_device);
}