#include "cuda_runtime.h"

template < unsigned block_size, typename T >
__device__ void red_kernel(unsigned tid, T* shared_buf)
{
    constexpr unsigned warp_size = 32;
    for (unsigned s = block_size / 2; s > warp_size; s >>= 1)
    {
        if (tid < s)
            shared_buf[tid] += shared_buf[tid + s];
        __syncthreads();
    }

    if (tid >= warp_size)
        return;

    // Warp reduce
    T my_sum = shared_buf[tid] + shared_buf[tid + warp_size];
    __syncwarp();
    shared_buf[tid] = my_sum;
    __syncwarp();
    my_sum += shared_buf[tid + 16];
    __syncwarp();
    shared_buf[tid] = my_sum;
    __syncwarp();
    my_sum += shared_buf[tid + 8];
    __syncwarp();
    shared_buf[tid] = my_sum;
    __syncwarp();
    my_sum += shared_buf[tid + 4];
    __syncwarp();
    shared_buf[tid] = my_sum;
    __syncwarp();
    my_sum += shared_buf[tid + 2];
    __syncwarp();
    shared_buf[tid] = my_sum;
    __syncwarp();
    my_sum += shared_buf[tid + 1];
    __syncwarp();
    if (tid == 0)
        shared_buf[0] = my_sum;
}