#include "cuda_runtime.h"

template < unsigned block_size, typename T >
__device__ void red_kernel(unsigned tid, T* shared_buf)
{
    constexpr unsigned warp_size = 32;
    for (unsigned s = block_size / 2; s > warp_size; s >>= 1)
    {
        if (tid < s)
            shared_buf[tid] += shared_buf[tid + s];
        __syncthreads();
    }

    if (tid >= warp_size)
        return;

    // Warp reduce
    constexpr unsigned all_mask = 0xffffffffu;
    T                  my_sum   = shared_buf[tid] + shared_buf[tid + warp_size];
    for (int offset = warp_size / 2; offset > 0; offset /= 2)
        my_sum += __shfl_down_sync(all_mask, my_sum, offset);
    if (tid == 0)
        shared_buf[0] = my_sum;
}