#include <cstdio>

#include "red_kernel5.cuh"

template < size_t block_size, size_t chunks_per_block, typename T >
__global__ void reduction(const T* data, size_t size, T* out)
{
    __shared__ T shared_buf[block_size];

    const unsigned tid = threadIdx.x;
    const unsigned bid = blockIdx.x;

    unsigned i      = bid * block_size * chunks_per_block + tid;
    shared_buf[tid] = (i < size) ? data[i] : 0.;
    for (unsigned chunk = 1; chunk < chunks_per_block; ++chunk)
    {
        i += block_size;
        if (i < size)
            shared_buf[tid] += data[i];
    }
    __syncthreads();
    red_kernel< block_size >(tid, shared_buf);
    if (tid == 0)
        out[bid] = shared_buf[0];
}